InstallMethod( ObjByExtRep,
        "for p-env family, list", true,
        [ IsPEnvElementFamily, IsList ], 0,
        function( fam, list )
    
    return Objectify( fam!.packedPElementDefaultType, [ Immutable( list ) ] );
end);


InstallMethod( PrintObj,
        "for penv elm", true,
        [ IsPEnvElement ], 0,
        function( elm )
    
    local   er,  i,  j;
    
    er:= elm![1];
    
    if er = [ ] then Print("<zero>"); fi;
    
    for i in [1,3..Length(er)-1] do
        if i > 1 then Print( "+" ); fi;
        Print("(",er[i+1],")*");
        if er[i] = [] then
            Print("1");
        else  
            for j in [1,3..Length(er[i])-1] do
                if j > 1 then Print("*"); fi;
                Print("x",er[i][j]);
                if er[i][j+1]>1 then Print("^",er[i][j+1]); fi;
            od;
        fi;
    od;
end );


#############################################################################
##
#M  OneOp( <m> ) . . . . . . . . . . . . . . . . for a PEnv element
#M  ZeroOp( <m> ) . . . . . . . . . . . . . . .  for a PEnv element
#M  \<( <m1>, <m2> ) . . . . . . . . . . . . . . for two PEnv elements
#M  \=( <m1>, <m2> ) . . . . . . . . . . . . . . for two PEnv elements
#M  \+( <m1>, <m2> ) . . . . . . . . . . . . . . for two PEnv elements
#M  \AINV( <m> )     . . . . . . . . . . . . . . for a PEnv element
#M  \in( <U>, <u> )  . . . . . . . . . . . . . . for PEnv, and element
##
InstallMethod( OneOp,
        "for PEnv element",
        true, [ IsPEnvElement and IsPackedElementDefaultRep ], 0,
        function( x )

    return ObjByExtRep( FamilyObj( x ), [ [], FamilyObj(x)!.oneField ] );

end );

InstallMethod( ZeroOp,
        "for PEnv element",
        true, [ IsPEnvElement and IsPackedElementDefaultRep ], 0,
        function( x )

    return ObjByExtRep( FamilyObj( x ), [ ] );

end );


InstallMethod( \<,
                "for two PEnv elements",
        IsIdenticalObj, [ IsPEnvElement and IsPackedElementDefaultRep,
                IsPEnvElement and IsPackedElementDefaultRep ], 0,
        function( x, y )
    return x![1]< y![1];
end );

InstallMethod( \=,
        "for two PEnv elements",
        IsIdenticalObj, [ IsPEnvElement and IsPackedElementDefaultRep,
                IsPEnvElement and IsPackedElementDefaultRep ], 0,
        function( x, y )

    return x![1] = y![1];
end );


InstallMethod( \+,
        "for two PEnv elements",
        true, [ IsPEnvElement and IsPackedElementDefaultRep,
                IsPEnvElement and IsPackedElementDefaultRep], 0,
        function( x, y )

    local   ex,  ey,  mons,  cfs,  i,  lst, len;
    
    # Insert one sorted list in the second one; 
    # can be done much more efficiently!
    
    ex:= x![1]; ey:= y![1];
    mons:= [ ]; cfs:= [ ];
    for i in [1,3..Length(ex)-1] do
        Add( mons, ex[i] ); Add( cfs, ex[i+1] );
    od;

    for i in [1,3..Length(ey)-1] do
        Add( mons, ey[i] ); Add( cfs, ey[i+1] );
    od;
    SortParallel( mons, cfs );
    lst:= [ ];
    for i in [1..Length( mons )] do
        len:= Length(lst);
        if len > 0 and lst[len-1] = mons[i] then
            lst[len]:= lst[len]+cfs[i];
            if lst[len] = 0*lst[len] then
                Unbind( lst[len-1] ); Unbind( lst[len] );
                lst:= Filtered( lst, x -> IsBound(x) );
            fi;

        else
            Add( lst, mons[i] ); Add( lst, cfs[i] );
        fi;
    od;
    return ObjByExtRep( FamilyObj(x), lst );
end );


InstallMethod( AINV,
        "for PEnv element",
        true, [ IsPEnvElement and IsPackedElementDefaultRep ], 0,
        function( x )

    local   ex,  i;

    ex:= ShallowCopy(x![1]);
    for i in [2,4..Length(ex)] do
        ex[i]:= -ex[i];
    od;
    return ObjByExtRep( FamilyObj(x), ex );
end );

InstallMethod( AINV_MUT,
        "for PEnv element",
        true, [ IsPEnvElement and IsPackedElementDefaultRep ], 0,
        function( x )

    local   ex,  i;

    ex:= ShallowCopy(x![1]);
    for i in [2,4..Length(ex)] do
        ex[i]:= -ex[i];
    od;
    return ObjByExtRep( FamilyObj(x), ex );
end );

#############################################################################
##
#M  \*( <scal>, <m> ) . . . . . . . . .for a scalar and a PEnv element
#M  \*( <m>, <scal> ) . . . . . . . . .for a scalar and a PEnv element
##
InstallMethod( \*,
        "for scalar and PEnv element",
        true, [ IsScalar, IsPEnvElement and
                IsPackedElementDefaultRep ], 0,
        function( scal, x )

    local   ex,  i;
    
    scal:= scal*FamilyObj( x )!.oneField;
    if IsZero( scal ) then return Zero(x); fi;
    ex:= ShallowCopy( x![1] );
    for i in [2,4..Length(ex)] do
        ex[i]:= scal*ex[i];
    od;
    return ObjByExtRep( FamilyObj(x), ex );
end);

InstallMethod( \*,
        "for PEnv element and scalar",
        true, [ IsPEnvElement and IsPackedElementDefaultRep,
                IsScalar ], 0,
        function( x, scal )

    local   ex,  i;
    
    scal:= scal*FamilyObj( x )!.oneField;
    if IsZero( scal ) then return Zero(x); fi;
    ex:= ShallowCopy( x![1] );
    for i in [2,4..Length(ex)] do
        ex[i]:= scal*ex[i];
    od;
    return ObjByExtRep( FamilyObj(x), ex );
end);

InstallMethod( \in,
        "for PEnv element and PEnv",
        true, [ IsPEnvElement, IsPEnvelopingAlgebra ], 0,
        function( u, U )
    
    return IsIdenticalObj( ElementsFamily( FamilyObj(U) ), FamilyObj(u) );
end );


InstallMethod( PrintObj,
        "for a p-enveloping algebra", true,
        [ IsPEnvelopingAlgebra ], 0,
        function( U )
    
    Print("PEnvAlgebra( ");
    ViewObj( UnderlyingLieAlgebra(U) );
    Print(" )");

end );

InstallMethod( ViewObj,
        "for a p-enveloping algebra", true,
        [ IsPEnvelopingAlgebra ], 0,
        function( U )
    
    Print("PEnvAlgebra( ");
    ViewObj( UnderlyingLieAlgebra(U) );
    Print(" )");
end );

InstallMethod( \*,
        "for two p-env elements", IsIdenticalObj,
        [ IsPEnvElement, IsPEnvElement ], 0,
        function( x, y )
    
    local   stretch,  P_reduction,  degs,  taglio,  ex,  ey,  list,  
            i,  j,  mon,  d,  k,  T,  pmap,  one,  p,  normalized,  
            len,  cf,  mm,  indices,  i2,  scalar,  head,  middle,  
            tail,  index,  Tcoeffs,  m,  l,  lennorm,  zero;
    
    stretch:= function( list )
        local res,j,k;
        
        res:= [];
        for j in [1,3..Length(list) - 1 ] do
            for k in [1..list[j+1]] do
                Append( res, [ list[j], 1 ] );
            od;
        od;
        return res;
    end;    

    P_reduction:= function( one, p, pmap, mon )
        
        local   m,  k,  j,  pol,  done,  cf,  found,  head,  tail,  
                im,  m1, t1;
        
        if mon=[] then return [mon,one]; fi;
        m:= [mon[1],mon[2]];
        k:= 1; 
        for j in [3,5..Length(mon)-1] do
            if mon[j] = m[k] then
                m[k+1]:= m[k+1]+mon[j+1];
            else
                Add( m, mon[j] );
                Add( m, mon[j+1] );
                k:= k+2;
            fi;
        od;
    
        pol:= [ m, one ];
        done:= 0;

        while done < Length(pol) do
            m := pol[done+1];
            cf:= pol[done+2];
            found:= false;
            for j in [2,4..Length(m)] do
                if m[j] >= p then
                    found:= true;
                    break;
                fi;
            od;
            if not found then 
                done:= done+2;
            else
                
                head:= m{[1..j-2]};
                if m[j] > p then
                    Add( head, m[j-1] );
                    Add( head, m[j]-p );
                fi;
                tail:= m{[j+1..Length(m)]};
                im:= pmap[m[j-1]];
                for k in [1,3..Length(im)-1] do
                    if head <> [] and im[k] = head[Length(head)-1] then
                        m1:= ShallowCopy( head );
                        m1[Length(m1)]:= m1[Length(m1)]+1;
                        Append( m1, tail );
                    elif Length(tail) > 0 and im[k] = tail[1] then
                        t1 := ShallowCopy( tail );
                        t1[2]:= t1[2] + 1;
                        m1:= ShallowCopy( head );
                        Append( m1, t1 );
                    else
                        m1:= ShallowCopy( head );
                        Append( m1, [im[k],1] );
                        Append( m1, tail );
                    fi;
                    Append( pol, [m1,cf*im[k+1]] );
                od;
                Unbind( pol[done+1]); Unbind(pol[done+2]);
                pol := Filtered( pol, x -> IsBound(x) );
            fi;
        od;
        
        for k in [1,3..Length(pol)-1] do
            pol[k]:= stretch( pol[k] );
        od;
        
        return pol;
        
    end;
    
    if IsBound( FamilyObj(x)!.degrees ) then
        degs:= FamilyObj(x)!.degrees;
        taglio:= true;
    else
        taglio:= false;
    fi;
    
    ex:= x![1]; ey:= y![1];
    list:= [ ];
    for i in [1,3..Length(ex)-1] do
        for j in [1,3..Length(ey)-1] do
            mon:= ShallowCopy( ex[i] );
            Append( mon, ey[j] );
            if taglio then
                d:= 0;
                for k in [1,3..Length(mon)-1] do
                    d:= d + mon[k+1]*degs[ mon[k] ];
                od;
                if d <= FamilyObj( x )!.maxDeg then
                    Add( list, mon );
                    Add( list, ex[i+1]*ey[j+1] );
                fi;
            else    
                Add( list, mon );
                Add( list, ex[i+1]*ey[j+1] );
            fi;
        od;
    od;
    
    T:= FamilyObj( x )!.sTab;
    pmap:= FamilyObj( x )!.pMap;
    one:= FamilyObj( x )!.oneField;
    p:= FamilyObj( x )!.charField;
    
    normalized := [];
    
    while not IsEmpty( list ) do
        
        list:= Compacted( list );
        len:= Length( list );
        for j in [1,3..len-1] do
            cf := list[j+1];
            mm:= P_reduction( one, p, pmap, list[j] );
            Unbind( list[j] );
            Unbind( list[j+1] );
            for k in [2,4..Length(mm)] do mm[k]:= mm[k]*cf; od;
            Append( list, mm );
        od;
        list:= Compacted(list);        
        
        # `indices' is a list of positive integers $[ j_1, j_2, \ldots, j_m ]$
        # s.t. the initial part $x_{i_1}^{e_1} \cdots x_{i_{j_k}}^{e_{j_k}}$
        # of the $k$-th monomial is known to be normalized,
        # i.e., $i_1 < i_2 < \cdots < i_{j_k}$.
        # (So $j_k = 1$ for all $k$ will always be correct.)
        indices:= ListWithIdenticalEntries( Length( list )/2, 1 );
        
        # Loop over the monomials that shall be normalized.
        for i in [ 1, 2 .. Length( indices ) ] do
            
            # If the `i'-th monomial is already normalized,
            # put it into `normalized'.
            # Otherwise swap the first non-ordered generators.
            i2:= 2*i;
            scalar:= list[ i2 ];
            mon:= list[ i2-1 ];
            len:= Length( mon );
            j:= 2 * indices[i] - 1;
            while j < len - 2 do
                
                if mon[j] < mon[ j+2 ] then
                    
                    # `mon' is better normalized than `indices' tells.
                    j:= j+2;
                    indices[i]:= indices[i] + 1;
                    
                elif mon[j] = mon[ j+2 ] then
                    
                    # absorption
                    mon[ j+1 ]:= mon[ j+1 ] + mon[ j+3 ];
                    for k in [ j+2 .. len-2 ] do
                        mon[k]:= mon[ k+2 ];
                    od;
                    Unbind( mon[  len  ] );
                    Unbind( mon[ len-1 ] );
                    len:= len - 2;
                    
                else
                    
                    # We must swap two generators.
                    # First construct head and tail of the arising monomials.
                    head:= mon{ [ 1 .. j-1 ] };
                    if 1 < mon[ j+1 ] then
                        Add( head, mon[  j  ]     );
                        Add( head, mon[ j+1 ] - 1 );
                    fi;
                    
                    middle:= [ mon[ j+2 ], 1, mon[j], 1 ];
                    
                    if 1 < mon[ j+3 ] then
                        tail:= Concatenation( [ mon[ j+2 ], mon[ j+3 ] - 1 ],
                                       mon{ [ j+4 .. len ] } );
                    else
                        tail:= mon{ [ j+4 .. len ] };
                    fi;
                    
                    # Adjust `indices[i]'.
                    index:= indices[i] - 1;
                    if index = 0 then
                        index:= 1;
                    fi;
                    
                    indices[i]:= index;
                    
                    # Replace the monomial by the swapped one.
                    list[ i2-1 ]:= Concatenation( head, middle, tail );
                    
                    # Add the coeffs/monomials that are given by the commutator.
                    # The part between `head' and `tail' of these list is
                    # $\sum_{k=1}^d c_{ijk} x_d$.
                    Tcoeffs:= T[ mon[j] ][ mon[ j+2 ] ];
                    for k in [ 1 .. Length( Tcoeffs[1] ) ] do
                        
                        m:= Concatenation( head, [ Tcoeffs[1][k], 1 ], tail );
                        cf:= scalar * Tcoeffs[2][k];
                        mm:= P_reduction( one, p, pmap, m );
                        for l in [2,4..Length(mm)] do mm[l]:= mm[l]*cf; od;
                        Append( list, mm );
                        Append( indices, List(mm,x->index) );
                    od;
                    
                    break;
                    
                fi;
                
            od;
            
            # If the monomial is normalized then move it to `normalized'.
            if len - 2 <= j then
                
                # Find the correct position in `normalized',
                # and insert the monomial.
                lennorm:= Length( normalized );
                k:= 2;
                while k <= lennorm do
                    if list[ i2-1 ] < normalized[ k-1 ] then
                        for l in [ lennorm, lennorm-1 .. k-1 ] do
                            normalized[l+2]:= normalized[l];
                        od;
                        normalized[ k-1 ]:= list[ i2-1 ];
                        normalized[  k  ]:= scalar;
                        break;
                    elif list[ i2-1 ] = normalized[ k-1 ] then
                        normalized[k]:= normalized[k] + scalar;
                        break;
                    fi;
                    k:= k+2;
                od;
                if lennorm < k then
                    normalized[ lennorm+1 ]:= list[ i2-1 ];
                    normalized[ lennorm+2 ]:= scalar;
                fi;
                
                # Remove the monomial from `list'.
                Unbind( list[ i2-1 ] );
                Unbind( list[  i2  ] );
                
            fi;
            
        od;
        
    od;
    
    # Remove monomials with multiplicity zero;
    if not IsEmpty( normalized ) then
        zero:= FamilyObj( x )!.zeroField;
        for i in [ 2, 4 .. Length( normalized ) ] do
            if normalized[i] = zero then
                Unbind( normalized[ i-1 ] );
                Unbind( normalized[  i  ] );
            fi;
        od;
        normalized:= Compacted( normalized );
    fi;
    
    # Return the normal form.
    return ObjByExtRep( FamilyObj(x), normalized );
    
end );


# Vector space functionality:
DeclareHandlingByNiceBasis( "IsSpaceOfPEnvElements",
    "for free left modules of elements of a p-enveloping algebra" );


#############################################################################
##
#M  NiceFreeLeftModuleInfo( <V> )
#M  NiceVector( <V>, <v> )
#M  UglyVector( <V>, <r> )
##
InstallHandlingByNiceBasis( "IsSpaceOfPEnvElements", rec(
    detect := function( F, gens, V, zero )
      return IsPEnvElementCollection( V );
      end,

    NiceFreeLeftModuleInfo := function( V )
      local gens,
            monomials,
            gen,
            list,
            zero,
            info;

      gens:= GeneratorsOfLeftModule( V );

      monomials:= [];

      for gen in gens do
        list:= gen![1];
        UniteSet( monomials, list{ [ 1, 3 .. Length( list ) - 1 ] } );
      od;

      zero:= Zero( LeftActingDomain( V ) );
      info:= rec( monomials := monomials,
                  zerocoeff := zero,
                  characteristic:= Characteristic( LeftActingDomain( V ) ), 
                  family    := ElementsFamily( FamilyObj( V ) ) );

      # For the zero row vector, catch the case of empty `monomials' list.
      if IsEmpty( monomials ) then
        info.zerovector := [ zero ];
      else
        info.zerovector := ListWithIdenticalEntries( Length( monomials ),
                                                     zero );
      fi;

      return info;
      end,

    NiceVector := function( V, v )
      local info, c, monomials, i, pos;
      info:= NiceFreeLeftModuleInfo( V );
      c:= ShallowCopy( info.zerovector );
      v:= v![1];
      monomials:= info.monomials;
      for i in [ 2, 4 .. Length( v ) ] do
        pos:= Position( monomials, v[ i-1 ] );
        if pos = fail then
          return fail;
        fi;
        c[ pos ]:= v[i];
      od;
      return c;
      end,

    UglyVector := function( V, r )
      local info, list, i;
      info:= NiceFreeLeftModuleInfo( V );
      if Length( r ) <> Length( info.zerovector ) then
        return fail;
      elif IsEmpty( info.monomials ) then
        if IsZero( r ) then
          return Zero( V );
        else
          return fail;
        fi;
      fi;
      list:= [];
      for i in [ 1 .. Length( r ) ] do
        if r[i] <> info.zerocoeff then
          Add( list, info.monomials[i] );
          Add( list, r[i] );
        fi;
      od;
      return ObjByExtRep( info.family, list  );
  end ) );
  


InstallMethod( PEnvelopingAlgebra,
        "for a restricted Lie algebra", true,
        [ IsLieAlgebra ], 0,
        function( L )
    
    local   taglio,  graded,  g,  h,  cls,  comps,  b,  BL,  degs,  i,  
            elmFam,  elmType,  F,  pm,  pmap,  j,  cf,  cf1,  k,  
            gens,  U,  tup,  bb,  grading,  t,  dd,  mon,  elm,  BU;
    
    if not IsRestrictedLieAlgebra( L ) then
        Error("<L> must be restricted");
    fi;
    
    taglio:= ValueOption( "cut" );
    
    if taglio = fail then taglio:= false; fi;

    # We get a basis of L; if L is graded then we take a basis
    # according to the grading.
    if HasGrading( L ) then
        graded:= true;
        g:= Grading( L );
        h:= g.hom_components;
        cls:= g.max_degree;
        comps:= List( [1..cls], i -> BasisVectors( Basis(h(i)) ) );        
        b:= Flat( comps );
        BL:= Basis( L, b );
        SetPthPowerImages( BL, List( b, x -> PthPowerImage( Basis(L), x ) ) );
        
        if taglio  then 
            degs:= [ ];
            for i in [1..Length(comps)] do
                Append( degs, List( comps[i], x -> i ) );
            od;
        fi;
        
    else
        if taglio  then
            Error("<L> must be graded");
        fi;
        graded:= false;
        BL:= Basis(L);
    fi;
    
    # Make famlily, type of the elements of U:
    elmFam:= NewFamily( "PEnvElmFam", IsPEnvElement );
    elmType:= NewType( elmFam, IsPackedElementDefaultRep );
    elmFam!.packedPElementDefaultType:= elmType;
    
    F:= LeftActingDomain( L );
    elmFam!.oneField:= One( F );
    elmFam!.zeroField:= Zero( F );
    elmFam!.charField:= Characteristic( F );
    if taglio then 
        elmFam!.degrees:= degs; 
        elmFam!.maxDeg:= g.max_degree;
    fi;
    
    pm:= PthPowerImages( BL );
    pmap:= [];
    for j in [1..Length(pm)] do
        cf:= Coefficients( BL, pm[j] );
        cf1:= [];
        for k in [1..Length(cf)] do
            if cf[k] <> Zero(F) then
                Append( cf1, [k,cf[k]] );
            fi;
        od;
        Add( pmap, cf1 );
    od;
    
    elmFam!.pMap:= pmap;
    elmFam!.sTab:= StructureConstantsTable( BL );
    
    # Make generators (starting with 1) 
    gens:= [ ObjByExtRep( elmFam, [ [], One(F) ] ) ]; 
    for i in [1..Length(BL)] do
        Add( gens, ObjByExtRep( elmFam, [ [ i, 1 ], One(F) ] ) );
    od;
    
    # Make U:
    U:= Objectify( NewType( CollectionsFamily( FamilyObj( gens[1] ) ),
                                IsMagmaRingModuloRelations
                            and IsPEnvelopingAlgebra
                            and IsAttributeStoringRep ),
                   rec() );
    SetIsAssociative( U, true );
    SetLeftActingDomain( U, F );
    SetGeneratorsOfLeftOperatorRing( U, gens );
    SetGeneratorsOfLeftOperatorRingWithOne( U, gens{[2..Length(gens)]} );
    SetOne( U, gens[1] );
    SetUnderlyingLieAlgebra( U, L );

    # If we cut U, then we also compute a basis of U:
    if taglio  then
        tup:= Tuples( [0..Characteristic( LeftActingDomain(L) )-1], 
                      Length(degs) );
        bb:= [ ];
        grading:= [ ];
        for t in tup do
            dd:= t*degs;
            if dd <= cls then
                mon:= [ ];
                for i in [1..Length(t)] do
                    if t[i]<>0 then
                        Add( mon, i ); Add( mon, t[i] );
                    fi;
                od;
                elm:= ObjByExtRep( elmFam, [ mon, One(F) ] );
                Add( bb, elm );
                if not IsBound( grading[dd+1] ) then
                    grading[dd+1]:= [ elm ];
                else
                    Add( grading[dd+1], elm );
                fi;
                
                
            fi;
        od;
        
        SetGrading( U, grading );
        
        BU:= Basis( U, bb );
        # The last command causes the program to compute a basis for U,
        # so the next command does not work...
        SetBasis( U, BU );
    fi;
    
    return U;
      
end );


annihilated_ideal:= function( A, g )
    
    local   B,  s,  eqs,  N,  k,  i,  c,  j,  sol;
    
    # g is a system of generators for A
    # we return the space of x\in A such that g*x=0
    # and x*g=0.
    
    B:= Basis(A);
    s:= Dimension( A );
    eqs:= [ ];
    N:= Length(g);
    
    eqs:= NullMat( 2*s*N, s, Zero( LeftActingDomain(A) ) ); 
    for k in [1..Length(g)] do
        for i in [1..s] do
            c:= Coefficients( B, g[k]*B[i] );
            for j in [1..Length(c)] do
                eqs[j+(k-1)*s][i]:= c[j];
            od;
        od;
    od;
    
    for k in [1..Length(g)] do
        for i in [1..s] do
            c:= Coefficients( B, B[i]*g[k] );
            for j in [1..Length(c)] do
                eqs[s*N+j+(k-1)*s][i]:= c[j];
            od;
        od;
    od;
    
    sol:= NullspaceMat( TransposedMat( eqs ) );
    return List( sol, x -> LinearCombination( x, B ) );
end;


InstallMethod( QuotientByMaximalIdeal,
        "for p-env algs", true,
        [ IsPEnvelopingAlgebra ], 0,
        
        function( U )
    
    local   F,  T,  A,  f,  g,  L,  ann,  V1,  V2,  W,  sp,  comp,  a,  
            I;
    
    F:= LeftActingDomain(U);
    T:= StructureConstantsTable( Basis(U) );
    A:= AlgebraByStructureConstants( F, T );
    f:= LeftModuleHomomorphismByImages( U, A, Basis(U), Basis(A) );
    g:= List( Grading(U)[2], x -> Image( f, x ) );
    L:= List( GeneratorsOfAlgebraWithOne(U), x -> Image( f, x ) );
    
    while true do
        ann:= annihilated_ideal( A, g );

        V1:= Subspace( A, ann );
        V2:= Subspace( A, L );
        W:= Intersection( V1, V2 );
        
        sp:= MutableBasis( F, L );
        comp:= [ ];
        for a in ann do
            if not IsContainedInSpan( sp, a ) then
                Add( comp, a );
                CloseMutableBasis( sp, a );
            fi;
        od;
        if comp = [ ] then
            return [ A, L ];
        fi;
        I:= Ideal( A, comp );
        f:= NaturalHomomorphismByIdeal( A, I );
        L:= List( L, x -> Image( f, x ) );
        g:= List( g, x -> Image( f, x ) );
        A:= Range( f );
    od;
end );

    
VSP:= function( p )
    
    local   F,  T,  L,  B,  gr,  i;
    
    F:= GF(p);
    T:= EmptySCTable(4,Zero(F),"antisymmetric");
    SetEntrySCTable( T, 1, 2, [ One(F), 3] );
    L:= LieAlgebraByStructureConstants( F, T );
    B:= Basis(L);
    SetPthPowerImages( B, [ Zero(L), Zero(L), B[4], Zero(L) ] );
    gr:= [ Subspace( L, [ B[1], B[2] ] ),
           Subspace( L, [ B[3] ] ) ];
    for i in [3..2*p-1] do gr[i]:= Subspace( L, [] ); od;
    gr[2*p]:= Subspace( L, [ B[4] ] );
    SetGrading( L, rec( min_degree:=1, 
                        max_degree:= 2*p,
                        source:= Integers,
                        hom_components:= function( d )
        
                               if d <= 2*p then 
                                  return gr[d];
                               else
                                  return Subspace( L, [] );
                               fi;
                           end ) );
     return L;
     
 end;
 
 
ESP:= function( p )
    
    local   F,  T,  L,  B,  gr,  i;
    
    F:= GF(p);
    T:= EmptySCTable(5,Zero(F),"antisymmetric");
    SetEntrySCTable( T, 1, 2, [ One(F), 3] );
    L:= LieAlgebraByStructureConstants( F, T );
    B:= Basis(L);
    SetPthPowerImages( B, [ Zero(L), Zero(L), B[4], B[5], Zero(L) ] );
    gr:= [ Subspace( L, [ B[1], B[2] ] ),
           Subspace( L, [ B[3] ] ) ];
    for i in [3..2*p^2-1] do gr[i]:= Subspace( L, [] ); od;
    gr[2*p]:= Subspace( L, [ B[4] ] );
    gr[2*p^2]:= Subspace( L, [ B[5] ] );
    SetGrading( L, rec( min_degree:=1, 
                        max_degree:= 2*p^2,
                        source:= Integers,
                        hom_components:= function( d )
        
                               if d <= 2*p^2 then 
                                  return gr[d];
                               else
                                  return Subspace( L, [] );
                               fi;
                           end ) );
     return L;
end; 
 
                           
table:= function( L )
    
    local   B,  d,  i,  j,  cf,  first,  k;
    
    B:= Basis(L);
    d:= Length(B);
    for i in [1..d] do
        for j in [i+1..d] do
            if not IsZero(B[i]*B[j]) then
                cf:= Coefficients( B, B[i]*B[j] );
                Print("[ x_",i,", x_",j," ] = ");
                first:= true;
                for k in [1..Length(cf)] do
                    if cf[k] <> 0*cf[k] then
                        if first then 
                            first:= false;
                        else
                            Print( " + " );
                        fi;
                        Print( IntFFE( cf[k] ), "x_",k );
                    fi;
                od;
                Print("\n");
            fi;
        od;
    od;
end;
