ExtendPLieAlgebra:= function( L, elms )
    
    local   B,  F,  S,  dimL,  dimK,  T,  i,  j,  K,  f,  pPow,  pos,  
            gr,  g,  max_deg,  sp,  vv,  d,  dnew,  max_degree,  
            source,  hom_components;
    
    B:= Basis(L);
    F:= LeftActingDomain(L);
    S:= StructureConstantsTable( B );
    dimL:= Dimension(L);
    dimK:= dimL+Length(elms);
    T:= EmptySCTable( dimK, Zero(F), "antisymmetric" );
    
    for i in [1..dimL] do
        for j in [1..dimL] do
            T[i][j]:= List( S[i][j], ShallowCopy );
        od;
    od;
    
    K:= LieAlgebraByStructureConstants( F, T );
    
    f:= LeftModuleHomomorphismByImages( L, K, B, Basis(K){[1..dimL]} );
    pPow:= List( PthPowerImages( B ), x -> Image( f,x ) );
    Append( pPow, List( elms, x -> Zero(K) ) );
    
    for i in [1..Length(elms)] do
        if not elms[i] in LieCentre(L) then
            Error("wrong");
        fi;
        if PthPowerImage( B, elms[i] ) <> Zero(L) then
            Error("wrong again");
        fi;
        pos:= Position( B, elms[i] );
        pPow[pos]:= Basis(K)[dimL+i];
    od;
    SetPthPowerImages( Basis(K), pPow );
    
    gr:= [ ];
    g:= Grading( L );
    max_deg:= g.max_degree;
    for i in [1..max_deg] do
        sp:= g.hom_components( i );
        vv:= List( Basis(sp), x -> Image( f, x ) );
        Add( gr, Subspace( K, vv ) );
    od;
    
    for i in [1..Length(elms)] do
        for j in [1..max_deg] do
            if elms[i] in g.hom_components(j) then
                d:= j;
                break;
            fi;
        od;
        dnew:= Characteristic(F)*d;
        if IsBound( gr[dnew] ) then
            vv:= ShallowCopy( BasisVectors( Basis( gr[dnew] ) ) );
            Add( vv, Basis(K)[dimL+i] );
            gr[dnew]:= Subspace( K, vv );
        else
            gr[dnew]:= Subspace( K, [Basis(K)[dimL+i]] );
        fi;
    od;
    max_deg:= Length(gr);
    for i in [1..max_deg] do
        if not IsBound( gr[i] ) then
            gr[i]:= Subspace( K, [] );
        fi;
    od;
    
    SetGrading( K, rec( min_degree:=1, 
                        max_degree:= max_deg,
                        source:= Integers,
                        hom_components:= function( d )
        
                               if d <= max_deg then 
                                  return gr[d];
                               else
                                  return Subspace( K, [] );
                               fi;
                           end ) );
                           
                           
    return K;
                           
    
end;
                       