DeclareRepresentation( "IsUEltDefaultRep", IsPositionalObjectRep,
        [ 1 ] );

DeclareCategory( "IsUElt", IsRingElement and 
        IsMultiplicativeElementWithInverse and
        IsAssociativeElement );

DeclareOperation( "UElement", [ IsRingElement ] );

DeclareOperation( "PCGroupByUElements", [ IsList ] );

##
InstallMethod( UElement,
        "for a ring element",
        true,
        [ IsRingElement ], 0,
        
        function( obj )
    
    local T, fam, ufam;
    
    fam:= FamilyObj( obj );
    if not IsBound( fam!.uEltDefFam ) then
       ufam:= NewFamily( "UEltFam", IsUElt );
       T:= NewType( ufam,  IsUElt and IsUEltDefaultRep );
       ufam!.defType:= T;
       fam!.uEltDefFam:= ufam;
    else
       T:= fam!.uEltDefFam!.defType;
    fi;
                           
    return Objectify( T, [ Immutable( obj ) ] );
end );


InstallMethod( PrintObj,
        "for a U element",
        true,
        [IsUElt], 0,
        function( obj )
    Print( "1 + ", obj![1] );
end );

InstallMethod( \*,
        "for u elements",
        IsIdenticalObj,
        [IsUElt and IsUEltDefaultRep,
         IsUElt and IsUEltDefaultRep ], 0,
        function(x,y)
    local x1,y1;
    x1:= x![1]; y1:= y![1];
    return UElement( x1+y1+x1*y1 );
end );

InstallMethod( Inverse,
        "for a U element",
        true,
        [IsUElt], 0,
        function( x )
    local x1,y,z,sign;
    
    x1:= x![1];
    y:= Zero( x1 );
    z:= x1;
    sign:= -1;
    while z <> 0*z do
        y := y + sign*z;
        z:= z*x1;
        sign:= -sign;
    od;
    return UElement( y );
end );

InstallMethod( INV,
        "for a U element",
        true,
        [IsUElt], 0,
        function( x )
    local x1,y,z,sign;
    
    x1:= x![1];
    y:= Zero( x1 );
    z:= x1;
    sign:= -1;
    while z <> 0*z do
        y := y + sign*z;
        z:= z*x1;
        sign:= -sign;
    od;
    return UElement( y );
end );

InstallMethod( One,
        "for a u element",
        true,
        [IsUElt], 0,
        x -> UElement( Zero( x![1] ) ) );


InstallMethod( \=,
        "for two u elements",
        IsIdenticalObj,
        [IsUElt and IsUEltDefaultRep,
         IsUElt and IsUEltDefaultRep], 0,
        function(x,y)
    return x![1]=y![1];
end );


InstallMethod( \<,
        "for two u elements",
        IsIdenticalObj,
        [IsUElt and IsUEltDefaultRep,
         IsUElt and IsUEltDefaultRep], 0,
        function(x,y)
    return x![1]<y![1];
    end );

    
InstallMethod( PCGroupByUElements,
        "for a list of uelements", true,
        [ IsList ], 0,
        function( list )
    
    local   G;
    
    G:= Group( list );
    Print("The size of G is ",Size(G),"\n");
    G:= Range( IsomorphismPermGroup( G ) );
    return Range( IsomorphismPcGroup(G) );
end );
